/* Copyright 2000 Red Hat, Inc.
 *
 * Florian La Roche <Florian.LaRoche@redhat.de>
 *
 * This software may be freely redistributed under the terms of the GNU
 * public license.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "rhutil.h"
#include <string.h>
#include <stdlib.h>

rh_list *
Buf2List (const char *buf)
{
	rh_list *list = NULL;
	const char *end;
	size_t n;

	if (!buf)
		return NULL;
	for (; *buf; buf = end) {
		if ((end = strchr (buf, '\n'))) {
			n = end - buf;
			/* allow "\r\n" newlines */
			if (n && end[-1] == '\r')
				n--;
			end++;
		} else {
			n = strlen (buf);
			end = buf + n;
		}
		list = rh_list_add (list, NULL, rh_xstrndup (buf, n), n);
	}
	return list;
}

static int length;

/* count size of each line plus trailing '\n' */
static int
count_length (rh_list *b)
{
	length += b->n + 1;
	return 0;
}

static char *buf;

static int
add_to_buf (rh_list *b)
{
	memcpy (buf, b->data, b->n);
	buf += b->n;
	*buf++ = '\n';
	return 0;
}

char *
List2Buf (rh_list *list, size_t *n)
{
	char *b;

	length = 0;
	(void) rh_list_traverse (&count_length, list, NULL);
	b = buf = rh_xmalloc (length + 1);
	(void) rh_list_traverse (&add_to_buf, list, NULL);
	*buf++ = '\0';
	if (n)
		*n = length;
	return b;
}
