/* Copyright 2000 Red Hat, Inc.
 *
 * Florian La Roche <Florian.LaRoche@redhat.de>
 *
 * This software may be freely redistributed under the terms of the GNU
 * public license.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "rhutil.h"
#include <stdlib.h>

static __inline__ rh_list *
rh_list_alloc (void)
{
	return rh_xcalloc (sizeof (rh_list));
}

void
rh_list_delete (rh_list *first, int free_data)
{
	rh_list *l = first;
	if (!first)
		return;
	do {
		rh_list *next = l->next;
		if (free_data) {
			RH_FREE (l->data);
		}
		free (l);
		l = next;
	} while (l != first);
}

rh_list *
rh_list_delete_node (rh_list *first, rh_list *pos)
{
	rh_list *ret;

	RH_FREE (pos->data);
	pos->prev->next = pos->next;
	pos->next->prev = pos->prev;
	ret = (pos == first) ? first->next : first;
	free (pos);
	return ret;
}

rh_list *
rh_list_add (rh_list *first, rh_list *pos, char *data, size_t n)
{
	rh_list *pnew = rh_list_alloc ();
	pnew->data = data;
	pnew->n = n;
	if (first) {
		/* append to the end if no position is given */
		if (!pos)
			pos = first;
		pnew->prev = pos->prev;
		pnew->next = pos;
		pos->prev->next = pnew;
		pos->prev = pnew;
	} else {
		pnew->prev = pnew;
		pnew->next = pnew;
		first = pnew;
	}
	return first;
}

rh_list *
rh_list_traverse (int (*func) (rh_list *), rh_list *first, rh_list *pos)
{
	if (!first)
		return NULL;
	if (!pos)
		pos = first;
	do {
		/* this allows to e.g. delete the current node. buahhh... */
		rh_list *next = pos->next;
		if ((*func) (pos))
			return pos;
		pos = next;
	} while (pos != first);
	return NULL;
}
