/*

	Datastructure-manipulation for the /proc items, XML based

	(C) 2000 Arjan van de Ven, licensed under the terms of the GPL
                                                                         
*/

#include <stdlib.h>
#include <stdio.h>
#include <parser.h>		/* libxml */
#include <string.h>
#include <assert.h>
#include <gtk/gtk.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>

#include <procxml.h>
#include <powertweak.h>

struct menuitem *xmlroot = NULL;

struct menuitem *findinlist(char *Item, struct menuitem *menu)
{
	if (Item == NULL)
		return NULL;
	if (menu == NULL)
		return NULL;

	while (menu != NULL) {
		if ((menu->Name!=NULL)&&(strcasecmp(Item, menu->Name) == 0))
			return menu;
		menu = menu->Next;
	}
	return NULL;
}

static struct menuitem *additemtolist(char *Item, struct menuitem **menu)
{
	struct menuitem *MenuItem;
	struct menuitem **Prev, *current;

	Prev = menu;
	current = *menu;
	
	if (Item==NULL)
		return NULL;


	while (current != NULL) {
		if ((current->Name!=NULL)&&(strcasecmp(Item, current->Name) == 0))
			return current;
		Prev = &(current->Next);
		current = current->Next;
	}

	MenuItem = malloc(sizeof(struct menuitem));
	if (MenuItem==NULL) {
		printf("Out of memory\n");
		exit(0);
	}
		
	memset(MenuItem, 0, sizeof(struct menuitem));
	MenuItem->Name = strdup(Item);
	MenuItem->Next = NULL;
	*Prev = MenuItem;

	return MenuItem;
}

void __add_xml_item(char *Menu, char *SubMenu, char *Tab, struct xml_item *xml_item, struct menuitem **rootmenu)
{
	struct menuitem *temp;
	
	if ((xml_item==NULL)||(xml_item->Name==NULL))
		return;

	if (Menu == NULL) {
		temp = additemtolist(xml_item->Name, rootmenu);
		temp->xml_item = xml_item;
	} else {
		temp = additemtolist(Menu, rootmenu);
		__add_xml_item(SubMenu, Tab, NULL, xml_item, &(temp->Submenus));
	}
}

void add_xml_item(struct xml_item *item, struct menuitem **rootmenu)
{
	assert(item != NULL);
	__add_xml_item(item->Menu, item->SubMenu, item->Tab, item, rootmenu);
}

struct xml_item *parse_item(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr cur)
{
	struct xml_item *xml_item = NULL;
	
	if (ns!=NULL) {}; /* Prevent a warning. "ns" is for future use */

	xml_item = malloc(sizeof(struct xml_item));
	if (xml_item == NULL) {
		printf("Out of memory\n");
		return NULL;
	}
	memset(xml_item, 0, sizeof(struct xml_item));

	cur = cur->childs;
	while (cur != NULL) {
		if (!strcasecmp(cur->name, "MENU"))
			xml_item->Menu = xmlNodeListGetString(doc, cur->childs, 1);
		if (!strcasecmp(cur->name, "SUBMENU"))
			xml_item->SubMenu = xmlNodeListGetString(doc, cur->childs, 1);
		if (!strcasecmp(cur->name, "TAB"))
			xml_item->Tab = xmlNodeListGetString(doc, cur->childs, 1);
		if (!strcasecmp(cur->name, "NAME"))
			xml_item->Name = xmlNodeListGetString(doc, cur->childs, 1);
		if (!strcasecmp(cur->name, "CONFIGNAME"))
			xml_item->ConfigName = xmlNodeListGetString(doc, cur->childs, 1);
		if (!strcasecmp(cur->name, "TYPE"))
			xml_item->Type = xmlNodeListGetString(doc, cur->childs, 1);
		if (!strcasecmp(cur->name, "FILE"))
			xml_item->File = xmlNodeListGetString(doc, cur->childs, 1);
		if (!strcasecmp(cur->name, "ELEMENT"))
			xml_item->Element = atoi(xmlNodeListGetString(doc, cur->childs, 1));
		if (!strcasecmp(cur->name, "LOW"))
			xml_item->MinValue = atoi(xmlNodeListGetString(doc, cur->childs, 1));
		if (!strcasecmp(cur->name, "HIGH"))
			xml_item->MaxValue = atoi(xmlNodeListGetString(doc, cur->childs, 1));
		if (!strcasecmp(cur->name, "DESCRIPTION"))
			xml_item->Description = xmlNodeListGetString(doc, cur->childs, 1);
		cur = cur->next;

	}

	return xml_item;
}


struct menuitem *parse_file(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr cur)
{
	struct menuitem *Root = NULL;
	
	if (cur==NULL)
		return NULL;
		
	cur = cur->childs;
	while (cur != NULL) {
		
		if ((cur->name!=NULL)&&(strcasecmp(cur->name, "PROCENTRY")==0)) {
			struct xml_item *item = parse_item(doc, ns, cur);
			if ((item != NULL) && ( item->File != NULL)) {
				if (fileexists(item->File)==TRUE)		
					add_xml_item(item , &Root);
			} else {
				add_xml_item(item , &Root);
			}
		}
		cur = cur->next;
	}

	return Root;
}


/* printRecurse is for debugging only */
void printRecurse(struct menuitem *menu, int depth)
{
	int i;

	while (menu != NULL) {
		for (i = 0; i < depth; i++)
			printf("  ");
		printf("%s", menu->Name);

		if (menu->xml_item != NULL)
			printf(" (%i,%i) ", menu->xml_item->MinValue, menu->xml_item->MaxValue);
			
		printf(" [%i] ",menu->ActiveCount);
		printf("\n");

		if (menu->Submenus != NULL)
			printRecurse(menu->Submenus, depth + 1);

		menu = menu->Next;
	}


}

/* 
 *	calculate_node_usage counts the number of "active" /proc items, ie. items
 *	that actually exist. It handles recursion in the obvious way:
 *      An node is assigned a count of "1" (if the /proc file exists) plus
 *      the sum of the nodes children.
 *
 * 	The return-value is the nodes value plus the value of all the ->Next neighbors
 *      so that you can say "  calculate_node_usage(menuitem->Submenus)" for the
 *      return-value to make sense.  
 *
 */
 
int calculate_node_usage(struct menuitem *menuitem)
{
	int temp;
	
	temp = 0;
	
	if (menuitem==NULL)
		return 0;
		
	if ((menuitem->xml_item!=NULL)&&(menuitem->xml_item->File!=NULL)) {
		if (fileexists(menuitem->xml_item->File)==TRUE)
			temp++;			
	}
	
	/* Count all child-nodes */
	temp += calculate_node_usage(menuitem->Submenus);
	
	/* Update the local field */
	menuitem->ActiveCount = temp;

	/* Traverse the list */
	temp +=	 calculate_node_usage(menuitem->Next);
	
	return temp;
	
}


void init_xml_subsystem_from_file(char *Filename)
{
	xmlDocPtr doc;
	xmlNodePtr cur;
	
	if (Filename==NULL)
		return;

	doc = xmlParseFile(Filename);
	if (doc == NULL) {
		printf("Severe XML error: doc == NULL!!");
		return;
	}

	cur = doc->root;
	if (cur == NULL) {
		printf("Severe XML error: cur == NULL");
		return;
	}

	xmlroot = parse_file(doc, NULL, cur);

	/* Calculate the actual number of /proc items present.
	   (void) because we don't care for the total, only the subtotals. */
	   
	(void)calculate_node_usage(xmlroot);
	
#ifdef DEBUG
	/* Debug code: print the tree to the screen */
	printRecurse(xmlroot, 0);
#endif


}
char *power2sysctl(char *str)
{
  char *nstr,*p;
  int len;

  if(!str)
     return 0;
   
  if(strstr(str, "/proc/sys/") == str)
    str += 10;

  nstr = strdup(str);

  if(!nstr)
    return 0;

  len = strlen(str);
  p = nstr;

  while(len) {
    if(*p == '/')
      *p = '.';
    p++;
    len--;
  }

  return nstr;
}


static  const char *key;
static  size_t keylen;
static char *value;
static  rh_list *first, *next;
static  size_t valuelen;


static int
find_data_start (const char *s)
{

  size_t i = 0;
  /* skip whitespace at the beginning of a line */
  while (isspace ((unsigned char)(s[i])))
    i++;
  /* check if the right key is in this line */
  if (strncmp (&s[i], key, keylen))
    return 0;
  i += keylen;

  /* skip whitespace */
  while (isspace ((unsigned char)(s[i])))
    i++;

  if (s[i] != '=')
    return 0;

  return i+1;		/* return beginning of the data */
}


static int
list_get_item (rh_list *l)
{
	char *s = l->data;
	size_t i = find_data_start (s);
	if (!i)
		return 0;
	value = string_trim (s + i, l->n - i, NULL);
	return 1;
}


char *
GetKeyShell (rh_list *first, rh_list **pos, const char *key2)
{
	rh_list *pos2;

	key = key2;
	keylen = strlen (key);
	value = NULL;

	pos2 = rh_list_traverse (&list_get_item, first, pos? *pos : NULL);
	if (pos)
		*pos = pos2;
	return value;
}



static rh_list *
SetShellKeyPos (rh_list *first, rh_list *pos,
	const char *key2, const char *value2)
{
	size_t key2len = strlen (key2);
	size_t value2len = strlen (value2);
	char *buf = rh_xmalloc (key2len + value2len + 4);
	memcpy (buf, key2, key2len);
	buf[key2len++] = ' ';
	buf[key2len++] = '=';
	buf[key2len] = ' ';
	memcpy (buf + key2len + 1, value2, value2len + 1);
	return rh_list_add (first, pos, buf, key2len + value2len + 1);
}

static int
list_set_item (rh_list *l)
{
	char *s = l->data;
	size_t i = find_data_start (s);
	if (!i)
		return 0;
	if (! value) {		/* line should be deleted */
		next = l->next;
		first = rh_list_delete_node (first, l);
	} else {		/* write new data */
		l->n = i + valuelen;
		s = l->data = rh_xrealloc (s, l->n + 1);
		memcpy (s + i, value, valuelen + 1);
	}
	return 1;
}

rh_list *
SetKeySysCtl (rh_list *first2, rh_list **pos2,
	const char *key2, const char *value2)
{
  rh_list *ret, *pos = pos2? *pos2 : NULL;
  
  key = key2;
  keylen = strlen (key);
  first = first2;
  valuelen = value2 ? strlen (value2) : 0;
  value = (char *)value2;

  if (! (ret = rh_list_traverse (&list_set_item, first, pos)) && value)
    /* append data if it wasn't in the list already */
    return SetShellKeyPos (first, NULL, key, value2);
  if (pos2)
    *pos2 = value2 ? ret : next;
  return first;
}

void save_XML(rh_list **rl, struct menuitem *menuitem)
{

  while (menuitem != NULL) {
    if (menuitem->xml_item != NULL) {

      char strbuf[1024];
      int len=sizeof(strbuf)-1;
      char *sysstr = power2sysctl(menuitem->xml_item->File);

      strbuf[1023]=0;
      
      if (strcmp(menuitem->xml_item->Type, "Checkbox") == 0)
	menuitem->xml_item->IntValue = menuitem->xml_item->BoolValue;
      
	if(!sysstr) {
	  fprintf(stderr, "No more memory!\n");
	  exit(10);
	}

	len -= snprintf(strbuf, len, "%d", menuitem->xml_item->IntValue);

	{
	  struct menuitem *next = menuitem->Next;

	  int element = 0;

	  while((next != NULL) 
		&& (next->xml_item) && (next->xml_item->Element > 0) 
		&& (0 == strcmp(menuitem->xml_item->File, 
				next->xml_item->File))
		&& (next->xml_item->Element > menuitem->xml_item->Element)
		&& (len > 0)) {
	    
	    menuitem = next;
	    element++;

	    while (element != -1 && element < menuitem->xml_item->Element 
		   && len > 0) {
	      int val = read_int_from_proc_file(menuitem->xml_item->File,
						element);
	      len -= snprintf(&strbuf[sizeof(strbuf)-len-1], len, " %d", val);
	      element++;
	    }

	    len -= snprintf(&strbuf[sizeof(strbuf)-len-1], len, " %d", 
			    next->xml_item->IntValue);
    
	    next = next->Next;
	  }
	}
	
	*rl = SetKeySysCtl(*rl, NULL, sysstr, strbuf);
	
	free(sysstr);      
    }

    if (menuitem->Submenus != NULL)
      save_XML(rl, menuitem->Submenus);
    
    menuitem = menuitem->Next;
  }

}

void init_XML_values(rh_list **rl, struct menuitem *menuitem)
{
   while (menuitem != NULL) {
      if (menuitem->xml_item != NULL) {
	 
	 char *sysstr = power2sysctl(menuitem->xml_item->File);
	 char *val;
	 
	 if (fileexists(menuitem->xml_item->File) == TRUE) {
	    if (strcmp(menuitem->xml_item->Type, "Spinbutton") == 0)
	      menuitem->xml_item->IntValue =
	      read_int_from_proc_file(menuitem->xml_item->File,
				      menuitem->xml_item->Element);
	    if (strcmp(menuitem->xml_item->Type, "Checkbox") == 0)
	      menuitem->xml_item->BoolValue =
	      read_int_from_proc_file(menuitem->xml_item->File,
				      menuitem->xml_item->Element);
	 }
	 val = GetKeyShell(*rl, NULL, sysstr);
	 if(val) {
	 	int element = menuitem->xml_item->Element;
		int i;
		if (element != -1) {
			for (i=0 ; i<element ; i++) {
				while (*val != '\t' && *val != ' ')
					val++;
				while (*val == '\t' || *val == ' ')
					val++;
			}
		}
	 
	    if (strcmp(menuitem->xml_item->Type, "Spinbutton") == 0)
	      menuitem->xml_item->IntValue = atoi(val);
	    if (strcmp(menuitem->xml_item->Type, "Checkbox") == 0)
	      menuitem->xml_item->BoolValue = atoi(val);
	 }
	}
      if (menuitem->Submenus != NULL)
	init_XML_values(rl, menuitem->Submenus);
      menuitem = menuitem->Next;
   }
}
